#!/usr/bin/perl

use strict;
use warnings;
use Fcntl qw(:seek);

print "XM2ZX16 CONVERTER\n";

my $infile = 'music.xm';
my $outfile = 'music.asm';
#my $debuglvl;

#pass dummy command line parameter if none present
my $debuglvl = $#ARGV + 1;
$ARGV[0] = '-0' if ($debuglvl == 0);

#check if music.xm is present, and open it if it is
if ( -e $infile ) {
	print "Converting...\n";
	open INFILE, $infile or die "Could not open $infile: $!";
	binmode INFILE;
} 
else {
	print "$infile not found\n";
	exit 1;
}

#create new music.asm
unlink $outfile if ( -e $outfile );
open OUTFILE, ">$outfile" or die $!;
print OUTFILE "\n\torg #8931\n";

#check if module has correct number of channels (16)
my $ix;
sysseek(INFILE, 68, 0) or die $!;
sysread(INFILE, $ix, 1) == 1 or die $!;
if ( ord($ix) != 16 ) {
	print "ERROR: Invalid number of channels in module\n";
	close INFILE;
	close OUTFILE;
	exit 1;
}

#determine song length
my ($jx, $songlength, $fileoffset, $patval);
sysseek(INFILE, 64, 0) or die $!;
sysread(INFILE, $songlength, 1) == 1 or die $!;
$songlength = ord($songlength);
print "song length:\t\t $songlength \n" if ( $ARGV[0] eq '-v' );

#determine number of unique patterns
my $uniqueptns;
sysseek(INFILE, 70, 0) or die $!;
sysread(INFILE, $uniqueptns, 1) == 1 or die $!;
$uniqueptns = ord($uniqueptns);
print "unique patterns:\t $uniqueptns \n" if ( $ARGV[0] eq '-v' );

#locate the pattern headers within the .xm source file and check pattern lengths
my (@ptnoffsetlist, @ptnlengths);
my ($headlength, $ptnlengthx, $packedlength, $plhibyte);

$ptnoffsetlist[0] = 336;
$fileoffset = $ptnoffsetlist[0];

for ($ix = 0; $ix < $uniqueptns; $ix++) {
	sysseek(INFILE, $fileoffset, 0) or die $!;	#read ptn header length
	sysread(INFILE, $headlength, 1) == 1 or die $!;
	$headlength = ord($headlength);
		
	$fileoffset = ($fileoffset) + 5;		#read ptn lengths
	sysseek(INFILE, $fileoffset, 0) or die $!;
	sysread(INFILE, $ptnlengthx, 1) == 1 or die $!;
	$ptnlengths[$ix] = ord($ptnlengthx);
		
	$fileoffset = ($fileoffset) + 2;		#read packed data length
	sysseek(INFILE, $fileoffset, 0) or die $!;
	sysread(INFILE, $packedlength, 1) == 1 or die $!;
	$packedlength = ord($packedlength);
	$fileoffset++;
	sysseek(INFILE, $fileoffset, 0) or die $!;
	sysread(INFILE, $plhibyte, 1) == 1 or die $!;
	$packedlength = $packedlength + ord($plhibyte)*256;

	$ptnoffsetlist[($ix+1)] = ($ptnoffsetlist[($ix)]) + ($headlength) + ($packedlength);
	print "pattern $ix starts at $ptnoffsetlist[$ix], length $ptnlengths[$ix] rows\n" if ( $ARGV[0] eq '-v' );
		
	$fileoffset = $fileoffset + $packedlength + 1;	#calculate pos of next ptn header
}

#set up default pattern hashtable
my (@ch1lst, @ch2lst, @ch3lst, @ch4lst, @ch5lst, @ch6lst, @ch7lst, @ch8lst, @ch9lst, @ch10lst, @ch11lst, @ch12lst, @ch13lst, @ch14lst, @ch15lst, @ch16lst);
my @ptnhashtable = (\@ch1lst,\@ch2lst,\@ch3lst,\@ch4lst,\@ch5lst,\@ch6lst,\@ch7lst,\@ch8lst,\@ch9lst,\@ch10lst,\@ch11lst,\@ch12lst,\@ch13lst,\@ch14lst,\@ch15lst,\@ch16lst);
my ($key,$val,$ref);

$fileoffset = 80;
for ($ix = 0; $ix <= $songlength-1; $ix++) {
	sysseek(INFILE, $fileoffset, 0) or die $!;
	sysread(INFILE, $patval, 1) == 1 or die $!;
	$patval = ord($patval);
	for ($jx = 1; $jx <= 16; $jx++) {
		$val = 'ch'.$jx.'ptn'.$patval;
		$ptnhashtable[$jx-1][$ix] = $val;
	}
	$fileoffset++;
}



#convert pattern data
my (@ch1, @ch2, @ch3, @ch4, @ch5, @ch6, @ch7, @ch8, @ch9, @ch10, @ch11, @ch12, @ch13, @ch14, @ch15, @ch16);
my @chanref = (\@ch1, \@ch2, \@ch3, \@ch4, \@ch5, \@ch6, \@ch7, \@ch8, \@ch9, \@ch10, \@ch11, \@ch12, \@ch13, \@ch14, \@ch15, \@ch16);
my ($rows, $cpval, $temp, $mx, $nx, $ptnusage, $ptnname, $xref);

#generate pattern name arrays
my (@ch1ptns, @ch2ptns, @ch3ptns, @ch4ptns, @ch5ptns, @ch6ptns, @ch7ptns, @ch8ptns, @ch9ptns, @ch10ptns, @ch11ptns, @ch12ptns, @ch13ptns, 
@ch14ptns, @ch15ptns, @ch16ptns);
my @ptnrefs = (\@ch1ptns, \@ch2ptns, \@ch3ptns, \@ch4ptns, \@ch5ptns, \@ch6ptns, \@ch7ptns, \@ch8ptns, \@ch9ptns, \@ch10ptns, \@ch11ptns, \@ch12ptns, \@ch13ptns, 
\@ch14ptns, \@ch15ptns, \@ch16ptns);



for ($ix = 0; $ix <= ($uniqueptns)-1; $ix++) {

	$ptnusage = IsPatternUsed($ix);

	if ($ptnusage == 1) {

		#print OUTFILE "ptn$ix\n";
		$ch1[0] = 97;	#tone
		$ch2[0] = 97;	#tone
		$ch3[0] = 97;	#tone
		$ch4[0] = 97;	#tone
		$ch5[0] = 97;	#tone
		$ch6[0] = 97;	#tone
		$ch7[0] = 97;	#tone
		$ch8[0] = 97;	#tone
		$ch9[0] = 97;	#tone
		$ch10[0] = 97;	#tone
		$ch11[0] = 97;	#tone
		$ch12[0] = 97;	#tone
		$ch13[0] = 97;	#tone
		$ch14[0] = 97;	#tone
		$ch15[0] = 97;	#tone
		$ch16[0] = 97;	#tone
	
		$fileoffset = $ptnoffsetlist[$ix] + 9;
	
		for ($rows = 1; $rows <= $ptnlengths[($ix)]; $rows++) {	#Achtung! Row values offset by -1 so we can preload dummy values

			$ch1[$rows] = $ch1[$rows-1];
			$ch2[$rows] = $ch2[$rows-1];
			$ch3[$rows] = $ch3[$rows-1];
			$ch4[$rows] = $ch4[$rows-1];
			$ch5[$rows] = $ch5[$rows-1];
			$ch6[$rows] = $ch6[$rows-1];
			$ch7[$rows] = $ch7[$rows-1];
			$ch8[$rows] = $ch8[$rows-1];
			$ch9[$rows] = $ch9[$rows-1];
			$ch10[$rows] = $ch10[$rows-1];
			$ch11[$rows] = $ch11[$rows-1];
			$ch12[$rows] = $ch12[$rows-1];
			$ch13[$rows] = $ch13[$rows-1];
			$ch14[$rows] = $ch14[$rows-1];
			$ch15[$rows] = $ch15[$rows-1];
			$ch16[$rows] = $ch16[$rows-1];
		
			for ($mx = 0; $mx <=15; $mx++) {			#reading 16 tracks per row
				sysseek(INFILE, $fileoffset, 0) or die $!;	#read control byte of row
				sysread(INFILE, $cpval, 1) == 1 or die $!;
				$cpval = ord($cpval);
		
				if ($cpval >= 128) {				#if we have compressed data
			
					$fileoffset++;
			
					if ($cpval != 128) {

						sysseek(INFILE, $fileoffset, 0) or die $!;	#read first data byte of row
						sysread(INFILE, $temp, 1) == 1 or die $!;
						$temp = ord($temp);
				
						if (($cpval&1) == 1) {				#if bit 0 is set, it's note -> counter val.		
							${$chanref[$mx]}[$rows] = $temp;			#set notes in ch1..16
							$fileoffset++;
						}
			
						$fileoffset++ if (($cpval&2) == 2);
						$fileoffset++ if (($cpval&4) == 4);
						$fileoffset++ if (($cpval&8) == 8);
						$fileoffset++ if (($cpval&16) == 16);
					}
				}
				else {			#if we have uncompressed data
					$temp = $cpval;
					${$chanref[$mx]}[$rows] = $temp;
					$fileoffset = $fileoffset + 5;
				}
			}
		}	
	}
	
	for ($nx = 0; $nx <= 15; $nx++) {		#copy temporary ptn data array to the permanent (anonymous) one
		@{$ptnrefs[$nx][$ix]} = @{$chanref[$nx]};	
	}
}


#replace duplicate patterns
my (@replacedptns, @temphash);
my ($patnametemp, $patnametemp2, $ox, $px, $qx, $rx, $temp2, $arrsize, $duplicate, $basearray, $cmparray);

for ($ix = 0; $ix < $uniqueptns; $ix++) {	#for each pattern
	for ($jx = 0; $jx <= 15; $jx++) {	#for each channel
	
		$temp = 0;
		$nx = $jx + 1;
		$patnametemp = 'ch'.$nx.'ptn'.$ix;
		
		foreach (@replacedptns) {	#check if pattern has already been replaced				
			$temp = 1 if ($_ eq $patnametemp);
		}
		if ($temp == 0) {		#if pattern isn't already on the "replaced patterns" list
			$px = $jx + 1;		#->set to 0 at the end of following outer for-loop
			$basearray = join(":",@{$ptnrefs[$jx][$ix]});
			
			for ($ox = $ix; $ox < $uniqueptns; $ox++) {
				for ($px = $jx; $px < 16; $px++) {
					$temp2 = 0;
					$rx = $px + 1;
					$patnametemp2 = 'ch'.$rx.'ptn'.$ox;
					foreach (@replacedptns) {
						$temp2 = 1 if ($_ eq $patnametemp2);
					}
					if ($temp2 == 0 && $patnametemp ne $patnametemp2) {
					
						$duplicate = 0;
						
						$cmparray = join(":",@{$ptnrefs[$px][$ox]});
						$duplicate = 1 if ($basearray eq $cmparray);
					
						if ($duplicate > 0) {
			
							push (@replacedptns,$patnametemp2);		#add pattern to @replacedpnts

							print "$patnametemp2 replaced by $patnametemp\n" if ( $ARGV[0] eq '-v' );
							
							for ($qx = 0; $qx < $songlength; $qx++) {
								$ptnhashtable[$px][$qx] = $patnametemp if ($ptnhashtable[$px][$qx] eq $patnametemp2);
							}
						}
					}
				}
				$px = 0;
			}
		
		}
	}
}

#setup note tables
my @notetabhi = (5,5,5,5,4,4,4,3,3,3,3,3,2,2,2,2,2,2,2,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0);

my @notetabmid = (249,163,82,5,189,121,57,252,195,141,90,41,252,209,169,130,94,60,28,254,225,198,173,148,126,104,84,65,47,30,14,255,240,227,214,202,191,180,170,
160,151,143,135,127,120,113,107,101,95,90,85,80,75,71,67,63,60,56,53,50,47,45,42,40,37,35,33,31,30,28,26,25,23,22,21,20,18,17,16,15,15,14,13,12,11);
      
my @notetablow = (13,0,7,173,152,122,21,103,53,23,19,217,105,128,3,215,185,172,138,52,142,140,10,245,60,199,136,107,97,90,69,26,202,70,133,122,28,99,68,180,175,
44,36,141,101,163,66,61,142,49,34,91,216,150,146,199,50,209,161,158,199,25,145,45,236,203,201,227,25,105,209,79,228,140,72,23,246,230,228,242,13,52,104,168,242);

my @keybdoffsets = (1,65,2,66,3,4,68,5,69,6,70,7,8,72,9,73,10,11,75,12,76,13,77,14,15,79,16,80,17,18,82,19,83,20,84,21,22,86,23,87,24,25,89,26,90,27,91,28,29,
93,30,94,31,32,96,33,97,34,98,35,36,100,37,101,38,39,103,40,104,41,105,42,43,107,44,108,45,46,110,47,111,48,112,49,50);

#print patterns
my ($tabbyte, $notelength, $patlengthremain);

#print "$ptnrefs[0][0]\n" if ( $ARGV[0] eq '-v' );	#debug
#print "$ptnrefs[0][1]\n\n" if ( $ARGV[0] eq '-v' );	#debug

for ($ix = 0; $ix < $uniqueptns; $ix++) {					#for each pattern
	
	$ptnusage = IsPatternUsed($ix);
	if ($ptnusage == 1) {
	
		for ($jx = 0; $jx < 16; $jx++) {				#for each channel
		
			$temp = $jx+1;
			$patnametemp = 'ch'.$temp.'ptn'.$ix;
			$temp = 0;
			foreach (@replacedptns) {					#check if pattern has been replaced
				$temp = 1 if ($_ eq $patnametemp);
			}
			
			if ($temp == 0) {
		
				$temp = $jx+1;
				print OUTFILE "ch$temp","ptn$ix\n";			#print pattern label

				for ($mx = 1; $mx <= $ptnlengths[$ix]; $mx++) {		#for each row
			
					$notelength = 1;				#set note length
					$tabbyte = $ptnrefs[$jx][$ix][$mx];		#convert xm note byte to zx16 lo/mid/hi bytes
				
											#determine note lengths and crunch patterns
					for ($patlengthremain = $ptnlengths[$ix] - $mx; $patlengthremain > 0; $patlengthremain--) {
						if ($tabbyte != $ptnrefs[$jx][$ix][$mx+1]) {
							last;
						}
						$notelength++;
						$mx++;
					}
				
				
					if ($tabbyte != 97) {
						print OUTFILE "\tdb #";
						printf OUTFILE "%x", $notetablow[$tabbyte];
						print OUTFILE ",#";
						printf OUTFILE "%x", $notetabmid[$tabbyte];
						print OUTFILE ",#";
						printf OUTFILE "%x", $notetabhi[$tabbyte];
						print OUTFILE ",#00,#";
						printf OUTFILE "%x", $notelength;
						print OUTFILE ",#";
						printf OUTFILE "%x", $keybdoffsets[$tabbyte];
						print OUTFILE "\n";
					}
					else {
						print OUTFILE "\tdb #00,#00,#00,#00,#";
						printf OUTFILE "%x", $notelength;
						print OUTFILE ",#00\n";
					}
					$notelength = 1;				#reset note length
				}
				print OUTFILE "\tdb #ff,#ff,#ff,#00,#00,#00\n\n";	#print pattern end
			}
			
		}
	}
}

#print pattern matrix
my $row;
for ($ix = 1; $ix <= 16; $ix++) {
	print OUTFILE "ch$ix\n\tdw ";
	print OUTFILE join("\n\tdw ",@{$ptnhashtable[$ix-1]});
	print OUTFILE "\n\tdw fend,#ffff\n\n";
}

#close files and exit
close INFILE;
close OUTFILE;

#check if a pattern is actually used
sub IsPatternUsed {
my ($fileoffset, $ptnval);
my $usage = 0;
my $patnum = $_[0];
	for ($fileoffset = 80; $fileoffset < ($songlength+80); $fileoffset++) {
	sysseek(INFILE, $fileoffset, 0) or die $!;
	sysread(INFILE, $ptnval, 1) == 1 or die $!;
	$usage = 1 if ($patnum == ord($ptnval));
	}
	return($usage);
}
